//Microsoft Developer Studio generated resource script.
//
#include "resource.h"

#define APSTUDIO_READONLY_SYMBOLS
/////////////////////////////////////////////////////////////////////////////
//
// Generated from the TEXTINCLUDE 2 resource.
//
#include <windows.h>
#ifndef IDC_STATIC
#define IDC_STATIC (-1)
#endif

/////////////////////////////////////////////////////////////////////////////
#undef APSTUDIO_READONLY_SYMBOLS

/////////////////////////////////////////////////////////////////////////////
// Neutral resources

#if !defined(AFX_RESOURCE_DLL) || defined(AFX_TARG_NEU)
#ifdef _WIN32
LANGUAGE LANG_NEUTRAL, SUBLANG_NEUTRAL
#pragma code_page(1252)
#endif //_WIN32

/////////////////////////////////////////////////////////////////////////////
//
// Dialog
//

IDD_RC_2PASS1 DIALOG DISCARDABLE  0, 0, 212, 212
STYLE DS_MODALFRAME | WS_POPUP | WS_CAPTION | WS_SYSMENU
CAPTION "1st Pass"
FONT 8, "MS Shell Dlg"
BEGIN
    EDITTEXT        IDC_STATS,72,6,106,12,ES_AUTOHSCROLL
    PUSHBUTTON      "...",IDC_STATS_BROWSE,182,7,22,12
    CONTROL         "Full quality first pass",IDC_FULL1PASS,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,15,38,83,12
    CONTROL         "Discard first pass",IDC_DISCARD1PASS,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,127,38,69,12
    LTEXT           "Stats filename:",IDC_STATIC,8,6,52,12,SS_CENTERIMAGE
    CTEXT           "If you don't discard first pass but keep full quality disabled,\nthe resulting 1st pass stream might not be mpeg-4 compliant.",
                    IDC_STATIC,7,114,197,35
    CTEXT           "Full quality first pass is only useful if you want to keep the resulting stream.\nIt doesn't improve quality of second pass and normally should be disabled.",
                    IDC_STATIC,7,71,197,35
END

IDD_MOTION DIALOG DISCARDABLE  0, 0, 212, 212
STYLE DS_MODALFRAME | WS_POPUP | WS_CAPTION | WS_SYSMENU
CAPTION "Motion"
FONT 8, "MS Shell Dlg"
BEGIN
    COMBOBOX        IDC_MOTION,112,19,76,76,CBS_DROPDOWNLIST | WS_VSCROLL | 
                    WS_TABSTOP
    COMBOBOX        IDC_VHQ,112,37,76,76,CBS_DROPDOWNLIST | WS_VSCROLL | 
                    WS_TABSTOP
    CONTROL         "Use chroma motion",IDC_CHROMAME,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,14,65,77,10
    CONTROL         "Turbo ;-)",IDC_TURBO,"Button",BS_AUTOCHECKBOX | 
                    WS_TABSTOP,143,65,43,10
    EDITTEXT        IDC_FRAMEDROP,112,96,75,12,ES_AUTOHSCROLL | ES_NUMBER
    EDITTEXT        IDC_MAXKEY,112,116,76,12,ES_AUTOHSCROLL | ES_NUMBER
    LTEXT           "Motion search precision:",IDC_STATIC,14,20,80,12,
                    SS_CENTERIMAGE
    LTEXT           "VHQ mode:",IDC_STATIC,14,39,38,8
    LTEXT           "Frame drop ratio:",IDC_FRAMEDROP_STATIC,14,96,68,12,
                    SS_CENTERIMAGE
    LTEXT           "Maximum I-frame interval:",IDC_STATIC,14,117,80,12,
                    SS_CENTERIMAGE
    GROUPBOX        "Motion Precision",IDC_STATIC,7,7,193,77
    GROUPBOX        "Other",IDC_STATIC,7,79,193,67
    CONTROL         "Use VHQ for bframes too",IDC_VHQ_BFRAME,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,14,52,95,10
END

IDD_MAIN DIALOGEX 0, 0, 225, 255
STYLE DS_MODALFRAME | WS_POPUP | WS_CAPTION | WS_SYSMENU
CAPTION "XviD Configuration"
FONT 8, "MS Shell Dlg"
BEGIN
    COMBOBOX        IDC_PROFILE,88,16,91,200,CBS_DROPDOWNLIST | WS_VSCROLL | 
                    WS_TABSTOP
    PUSHBUTTON      "more...",IDC_PROFILE_ADV,184,16,28,12
    COMBOBOX        IDC_MODE,88,34,92,76,CBS_DROPDOWNLIST | WS_VSCROLL | 
                    WS_TABSTOP
    PUSHBUTTON      "more...",IDC_MODE_ADV,184,34,28,12
    CTEXT           "Target bitrate (kbps):",IDC_BITRATE_S,12,53,71,13,
                    SS_NOTIFY | SS_CENTERIMAGE,WS_EX_DLGMODALFRAME
    EDITTEXT        IDC_BITRATE,88,53,91,12,ES_AUTOHSCROLL
    PUSHBUTTON      "calc...",IDC_BITRATE_ADV,184,52,28,12
    CONTROL         "Slider1",IDC_SLIDER,"msctls_trackbar32",TBS_BOTH | 
                    TBS_NOTICKS | WS_TABSTOP,12,78,204,14
    CONTROL         "List1",IDC_ZONES,"SysListView32",LVS_REPORT | 
                    LVS_SINGLESEL | LVS_SHOWSELALWAYS | LVS_NOSORTHEADER | 
                    WS_TABSTOP,14,110,198,68,WS_EX_STATICEDGE
    PUSHBUTTON      "Add",IDC_ADD,16,182,36,12
    PUSHBUTTON      "Remove",IDC_REMOVE,56,182,36,12
    PUSHBUTTON      "Zone Options...",IDC_EDIT,155,182,57,12
    PUSHBUTTON      "Load Defaults",IDC_DEFAULTS,7,238,64,13
    PUSHBUTTON      "Other Options...",IDC_OTHER,81,238,64,13
    DEFPUSHBUTTON   "OK",IDOK,156,238,64,13
    GROUPBOX        "Main Settings",IDC_STATIC,7,3,212,202
    LTEXT           "Encoding type:",IDC_STATIC,14,34,70,12,SS_CENTERIMAGE
    LTEXT           "Profile @ Level:",IDC_STATIC,14,16,70,12,SS_CENTERIMAGE
    LTEXT           "X",IDC_BITRATE_MIN,16,70,79,8
    RTEXT           "X",IDC_BITRATE_MAX,125,70,84,8
    GROUPBOX        "Zones",IDC_STATIC,7,98,212,107
    GROUPBOX        "More",IDC_STATIC,7,200,212,29
    COMBOBOX        IDC_QUALITY,86,210,92,76,CBS_DROPDOWNLIST | WS_VSCROLL | 
                    WS_TABSTOP
    PUSHBUTTON      "more...",IDC_QUALITY_ADV,182,210,28,12
    LTEXT           "Quality preset:",IDC_STATIC,12,210,70,12,SS_CENTERIMAGE
END

IDD_QUANT DIALOG DISCARDABLE  0, 0, 212, 212
STYLE DS_MODALFRAME | WS_POPUP | WS_CAPTION | WS_SYSMENU
CAPTION "Quantization"
FONT 8, "MS Shell Dlg"
BEGIN
    EDITTEXT        IDC_MINIQUANT,120,18,76,12,ES_AUTOHSCROLL | ES_NUMBER
    EDITTEXT        IDC_MAXIQUANT,120,34,76,12,ES_AUTOHSCROLL | ES_NUMBER
    EDITTEXT        IDC_MINPQUANT,120,50,76,12,ES_AUTOHSCROLL | ES_NUMBER
    EDITTEXT        IDC_MAXPQUANT,120,66,76,12,ES_AUTOHSCROLL | ES_NUMBER
    EDITTEXT        IDC_MINBQUANT,120,82,76,12,ES_AUTOHSCROLL | ES_NUMBER
    EDITTEXT        IDC_MAXBQUANT,120,98,76,12,ES_AUTOHSCROLL | ES_NUMBER
    GROUPBOX        "Quantizer restrictions",IDC_STATIC,8,6,196,112
    LTEXT           "Min I-frame quantizer:",IDC_STATIC,16,18,76,12,
                    SS_CENTERIMAGE
    LTEXT           "Max I-frame quantizer:",IDC_STATIC,16,34,76,12,
                    SS_CENTERIMAGE
    LTEXT           "Min P-frame quantizer:",IDC_STATIC,16,50,76,12,
                    SS_CENTERIMAGE
    LTEXT           "Max P-frame quantizer:",IDC_STATIC,16,66,76,12,
                    SS_CENTERIMAGE
    LTEXT           "Min B-frame quantizer:",IDC_STATIC,16,82,76,12,
                    SS_CENTERIMAGE
    LTEXT           "Max B-frame quantizer:",IDC_STATIC,16,98,76,12,
                    SS_CENTERIMAGE
    CONTROL         "Trellis quantization",IDC_TRELLISQUANT,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,8,128,74,10
END

IDD_RC_2PASS2 DIALOG DISCARDABLE  0, 0, 212, 212
STYLE DS_MODALFRAME | WS_POPUP | WS_CAPTION | WS_SYSMENU
CAPTION "2nd Pass"
FONT 8, "MS Shell Dlg"
BEGIN
    EDITTEXT        IDC_STATS,72,6,112,12,ES_AUTOHSCROLL
    PUSHBUTTON      "...",IDC_STATS_BROWSE,189,7,15,11
    EDITTEXT        IDC_KFBOOST,140,34,56,12,ES_AUTOHSCROLL | ES_NUMBER
    EDITTEXT        IDC_MINKEY,140,55,56,12,ES_AUTOHSCROLL | ES_NUMBER
    EDITTEXT        IDC_KFREDUCTION,140,69,56,12,ES_AUTOHSCROLL | ES_NUMBER
    EDITTEXT        IDC_OVERFLOW_CONTROL_STRENGTH,140,99,56,12,
                    ES_AUTOHSCROLL | ES_NUMBER
    EDITTEXT        IDC_OVERIMP,140,118,56,12,ES_AUTOHSCROLL | ES_NUMBER
    EDITTEXT        IDC_OVERDEG,140,137,56,12,ES_AUTOHSCROLL | ES_NUMBER
    EDITTEXT        IDC_CURVECOMPH,140,169,56,12,ES_AUTOHSCROLL | ES_NUMBER
    EDITTEXT        IDC_CURVECOMPL,140,186,56,12,ES_AUTOHSCROLL | ES_NUMBER
    LTEXT           "Stats filename:",IDC_STATIC,8,6,52,12,SS_CENTERIMAGE
    GROUPBOX        "Intra-frames tuning",IDC_STATIC,7,22,197,69
    LTEXT           "Overflow control strength (%):",IDC_STATIC,16,100,100,
                    12,SS_CENTERIMAGE
    LTEXT           "High bitrate scenes degradation (%):",IDC_STATIC,16,170,
                    124,12,SS_CENTERIMAGE
    LTEXT           "Low bitrate scenes improvement (%):",IDC_STATIC,16,186,
                    124,12,SS_CENTERIMAGE
    LTEXT           "I-frame boost (%):",IDC_STATIC,16,34,91,12,
                    SS_CENTERIMAGE
    LTEXT           "...are reduced by (%):",IDC_STATIC,16,69,100,12
    LTEXT           "Max overflow improvement (%):",IDC_STATIC,16,119,100,12,
                    SS_CENTERIMAGE
    LTEXT           "Max overflow degradation (%):",IDC_STATIC,16,138,100,12,
                    SS_CENTERIMAGE
    LTEXT           "I-frames closer than... (frames):",IDC_STATIC,16,56,124,
                    11
    GROUPBOX        "Overflow treatment",IDC_STATIC,7,86,197,75
    GROUPBOX        "Curve compression",IDC_STATIC,7,156,197,51
END

IDD_ENC DIALOG DISCARDABLE  0, 0, 212, 212
STYLE DS_MODALFRAME | WS_POPUP | WS_CAPTION | WS_SYSMENU
CAPTION "Encoder"
FONT 8, "MS Shell Dlg"
BEGIN
    EDITTEXT        IDC_NUMTHREADS,127,32,75,12,ES_AUTOHSCROLL | ES_NUMBER
    COMBOBOX        IDC_FOURCC,127,13,76,76,CBS_DROPDOWNLIST | WS_VSCROLL | 
                    WS_TABSTOP
    CONTROL         "Print debug info on each frame",IDC_VOPDEBUG,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,7,51,113,10
    CONTROL         "Display encoding status",IDC_DISPLAY_STATUS,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,7,64,91,10
    LTEXT           "Number of threads:",IDC_NUMTHREADS_STATIC,7,34,61,8
    LTEXT           "FourCC used:",IDC_STATIC,7,15,80,8,SS_CENTERIMAGE
END

IDD_QUANTMATRIX DIALOG DISCARDABLE  0, 0, 288, 149
STYLE DS_MODALFRAME | WS_POPUP | WS_CAPTION | WS_SYSMENU
CAPTION "Custom quantization matrix"
FONT 8, "MS Shell Dlg"
BEGIN
    DEFPUSHBUTTON   "OK",IDOK,172,128,47,13
    EDITTEXT        IDC_QINTRA00,8,16,15,11,ES_AUTOHSCROLL | ES_NUMBER | 
                    WS_DISABLED | NOT WS_BORDER
    EDITTEXT        IDC_QINTRA01,24,16,15,11,ES_AUTOHSCROLL | ES_NUMBER | 
                    NOT WS_BORDER
    EDITTEXT        IDC_QINTRA02,40,16,15,11,ES_AUTOHSCROLL | ES_NUMBER | 
                    NOT WS_BORDER
    EDITTEXT        IDC_QINTRA03,56,16,15,11,ES_AUTOHSCROLL | ES_NUMBER | 
                    NOT WS_BORDER
    EDITTEXT        IDC_QINTRA04,72,16,15,11,ES_AUTOHSCROLL | ES_NUMBER | 
                    NOT WS_BORDER
    EDITTEXT        IDC_QINTRA05,88,16,15,11,ES_AUTOHSCROLL | ES_NUMBER | 
                    NOT WS_BORDER
    EDITTEXT        IDC_QINTRA06,104,16,15,11,ES_AUTOHSCROLL | ES_NUMBER | 
                    NOT WS_BORDER
    EDITTEXT        IDC_QINTRA07,120,16,15,11,ES_AUTOHSCROLL | ES_NUMBER | 
                    NOT WS_BORDER
    EDITTEXT        IDC_QINTRA08,8,28,15,11,ES_AUTOHSCROLL | ES_NUMBER | NOT 
                    WS_BORDER
    EDITTEXT        IDC_QINTRA09,24,28,15,11,ES_AUTOHSCROLL | ES_NUMBER | 
                    NOT WS_BORDER
    EDITTEXT        IDC_QINTRA10,40,28,15,11,ES_AUTOHSCROLL | ES_NUMBER | 
                    NOT WS_BORDER
    EDITTEXT        IDC_QINTRA11,56,28,15,11,ES_AUTOHSCROLL | ES_NUMBER | 
                    NOT WS_BORDER
    EDITTEXT        IDC_QINTRA12,72,28,15,11,ES_AUTOHSCROLL | ES_NUMBER | 
                    NOT WS_BORDER
    EDITTEXT        IDC_QINTRA13,88,28,15,11,ES_AUTOHSCROLL | ES_NUMBER | 
                    NOT WS_BORDER
    EDITTEXT        IDC_QINTRA14,104,28,15,11,ES_AUTOHSCROLL | ES_NUMBER | 
                    NOT WS_BORDER
    EDITTEXT        IDC_QINTRA15,120,28,15,11,ES_AUTOHSCROLL | ES_NUMBER | 
                    NOT WS_BORDER
    EDITTEXT        IDC_QINTRA16,8,40,15,11,ES_AUTOHSCROLL | ES_NUMBER | NOT 
                    WS_BORDER
    EDITTEXT        IDC_QINTRA17,24,40,15,11,ES_AUTOHSCROLL | ES_NUMBER | 
                    NOT WS_BORDER
    EDITTEXT        IDC_QINTRA18,40,40,15,11,ES_AUTOHSCROLL | ES_NUMBER | 
                    NOT WS_BORDER
    EDITTEXT        IDC_QINTRA19,56,40,15,11,ES_AUTOHSCROLL | ES_NUMBER | 
                    NOT WS_BORDER
    EDITTEXT        IDC_QINTRA20,72,40,15,11,ES_AUTOHSCROLL | ES_NUMBER | 
                    NOT WS_BORDER
    EDITTEXT        IDC_QINTRA21,88,40,15,11,ES_AUTOHSCROLL | ES_NUMBER | 
                    NOT WS_BORDER
    EDITTEXT        IDC_QINTRA22,104,40,15,11,ES_AUTOHSCROLL | ES_NUMBER | 
                    NOT WS_BORDER
    EDITTEXT        IDC_QINTRA23,120,40,15,11,ES_AUTOHSCROLL | ES_NUMBER | 
                    NOT WS_BORDER
    EDITTEXT        IDC_QINTRA24,8,52,15,11,ES_AUTOHSCROLL | ES_NUMBER | NOT 
                    WS_BORDER
    EDITTEXT        IDC_QINTRA25,24,52,15,11,ES_AUTOHSCROLL | ES_NUMBER | 
                    NOT WS_BORDER
    EDITTEXT        IDC_QINTRA26,40,52,15,11,ES_AUTOHSCROLL | ES_NUMBER | 
                    NOT WS_BORDER
    EDITTEXT        IDC_QINTRA27,56,52,15,11,ES_AUTOHSCROLL | ES_NUMBER | 
                    NOT WS_BORDER
    EDITTEXT        IDC_QINTRA28,72,52,15,11,ES_AUTOHSCROLL | ES_NUMBER | 
                    NOT WS_BORDER
    EDITTEXT        IDC_QINTRA29,88,52,15,11,ES_AUTOHSCROLL | ES_NUMBER | 
                    NOT WS_BORDER
    EDITTEXT        IDC_QINTRA30,104,52,15,11,ES_AUTOHSCROLL | ES_NUMBER | 
                    NOT WS_BORDER
    EDITTEXT        IDC_QINTRA31,120,52,15,11,ES_AUTOHSCROLL | ES_NUMBER | 
                    NOT WS_BORDER
    EDITTEXT        IDC_QINTRA32,8,64,15,11,ES_AUTOHSCROLL | ES_NUMBER | NOT 
                    WS_BORDER
    EDITTEXT        IDC_QINTRA33,24,64,15,11,ES_AUTOHSCROLL | ES_NUMBER | 
                    NOT WS_BORDER
    EDITTEXT        IDC_QINTRA34,40,64,15,11,ES_AUTOHSCROLL | ES_NUMBER | 
                    NOT WS_BORDER
    EDITTEXT        IDC_QINTRA35,56,64,15,11,ES_AUTOHSCROLL | ES_NUMBER | 
                    NOT WS_BORDER
    EDITTEXT        IDC_QINTRA36,72,64,15,11,ES_AUTOHSCROLL | ES_NUMBER | 
                    NOT WS_BORDER
    EDITTEXT        IDC_QINTRA37,88,64,15,11,ES_AUTOHSCROLL | ES_NUMBER | 
                    NOT WS_BORDER
    EDITTEXT        IDC_QINTRA38,104,64,15,11,ES_AUTOHSCROLL | ES_NUMBER | 
                    NOT WS_BORDER
    EDITTEXT        IDC_QINTRA39,120,64,15,11,ES_AUTOHSCROLL | ES_NUMBER | 
                    NOT WS_BORDER
    EDITTEXT        IDC_QINTRA40,8,76,15,11,ES_AUTOHSCROLL | ES_NUMBER | NOT 
                    WS_BORDER
    EDITTEXT        IDC_QINTRA41,24,76,15,11,ES_AUTOHSCROLL | ES_NUMBER | 
                    NOT WS_BORDER
    EDITTEXT        IDC_QINTRA42,40,76,15,11,ES_AUTOHSCROLL | ES_NUMBER | 
                    NOT WS_BORDER
    EDITTEXT        IDC_QINTRA43,56,76,15,11,ES_AUTOHSCROLL | ES_NUMBER | 
                    NOT WS_BORDER
    EDITTEXT        IDC_QINTRA44,72,76,15,11,ES_AUTOHSCROLL | ES_NUMBER | 
                    NOT WS_BORDER
    EDITTEXT        IDC_QINTRA45,88,76,15,11,ES_AUTOHSCROLL | ES_NUMBER | 
                    NOT WS_BORDER
    EDITTEXT        IDC_QINTRA46,104,76,15,11,ES_AUTOHSCROLL | ES_NUMBER | 
                    NOT WS_BORDER
    EDITTEXT        IDC_QINTRA47,120,76,15,11,ES_AUTOHSCROLL | ES_NUMBER | 
                    NOT WS_BORDER
    EDITTEXT        IDC_QINTRA48,8,88,15,11,ES_AUTOHSCROLL | ES_NUMBER | NOT 
                    WS_BORDER
    EDITTEXT        IDC_QINTRA49,24,88,15,11,ES_AUTOHSCROLL | ES_NUMBER | 
                    NOT WS_BORDER
    EDITTEXT        IDC_QINTRA50,40,88,15,11,ES_AUTOHSCROLL | ES_NUMBER | 
                    NOT WS_BORDER
    EDITTEXT        IDC_QINTRA51,56,88,15,11,ES_AUTOHSCROLL | ES_NUMBER | 
                    NOT WS_BORDER
    EDITTEXT        IDC_QINTRA52,72,88,15,11,ES_AUTOHSCROLL | ES_NUMBER | 
                    NOT WS_BORDER
    EDITTEXT        IDC_QINTRA53,88,88,15,11,ES_AUTOHSCROLL | ES_NUMBER | 
                    NOT WS_BORDER
    EDITTEXT        IDC_QINTRA54,104,88,15,11,ES_AUTOHSCROLL | ES_NUMBER | 
                    NOT WS_BORDER
    EDITTEXT        IDC_QINTRA55,120,88,15,11,ES_AUTOHSCROLL | ES_NUMBER | 
                    NOT WS_BORDER
    EDITTEXT        IDC_QINTRA56,8,100,15,11,ES_AUTOHSCROLL | ES_NUMBER | 
                    NOT WS_BORDER
    EDITTEXT        IDC_QINTRA57,24,100,15,11,ES_AUTOHSCROLL | ES_NUMBER | 
                    NOT WS_BORDER
    EDITTEXT        IDC_QINTRA58,40,100,15,11,ES_AUTOHSCROLL | ES_NUMBER | 
                    NOT WS_BORDER
    EDITTEXT        IDC_QINTRA59,56,100,15,11,ES_AUTOHSCROLL | ES_NUMBER | 
                    NOT WS_BORDER
    EDITTEXT        IDC_QINTRA60,72,100,15,11,ES_AUTOHSCROLL | ES_NUMBER | 
                    NOT WS_BORDER
    EDITTEXT        IDC_QINTRA61,88,100,15,11,ES_AUTOHSCROLL | ES_NUMBER | 
                    NOT WS_BORDER
    EDITTEXT        IDC_QINTRA62,104,100,15,11,ES_AUTOHSCROLL | ES_NUMBER | 
                    NOT WS_BORDER
    EDITTEXT        IDC_QINTRA63,120,100,15,11,ES_AUTOHSCROLL | ES_NUMBER | 
                    NOT WS_BORDER
    GROUPBOX        "Intra matrix",IDC_STATIC,4,4,136,112
    EDITTEXT        IDC_QINTER00,152,16,15,11,ES_AUTOHSCROLL | ES_NUMBER | 
                    NOT WS_BORDER
    EDITTEXT        IDC_QINTER01,168,16,15,11,ES_AUTOHSCROLL | ES_NUMBER | 
                    NOT WS_BORDER
    EDITTEXT        IDC_QINTER02,184,16,15,11,ES_AUTOHSCROLL | ES_NUMBER | 
                    NOT WS_BORDER
    EDITTEXT        IDC_QINTER03,200,16,15,11,ES_AUTOHSCROLL | ES_NUMBER | 
                    NOT WS_BORDER
    EDITTEXT        IDC_QINTER04,216,16,15,11,ES_AUTOHSCROLL | ES_NUMBER | 
                    NOT WS_BORDER
    EDITTEXT        IDC_QINTER05,232,16,15,11,ES_AUTOHSCROLL | ES_NUMBER | 
                    NOT WS_BORDER
    EDITTEXT        IDC_QINTER06,248,16,15,11,ES_AUTOHSCROLL | ES_NUMBER | 
                    NOT WS_BORDER
    EDITTEXT        IDC_QINTER07,264,16,15,11,ES_AUTOHSCROLL | ES_NUMBER | 
                    NOT WS_BORDER
    EDITTEXT        IDC_QINTER08,152,28,15,11,ES_AUTOHSCROLL | ES_NUMBER | 
                    NOT WS_BORDER
    EDITTEXT        IDC_QINTER09,168,28,15,11,ES_AUTOHSCROLL | ES_NUMBER | 
                    NOT WS_BORDER
    EDITTEXT        IDC_QINTER10,184,28,15,11,ES_AUTOHSCROLL | ES_NUMBER | 
                    NOT WS_BORDER
    EDITTEXT        IDC_QINTER11,200,28,15,11,ES_AUTOHSCROLL | ES_NUMBER | 
                    NOT WS_BORDER
    EDITTEXT        IDC_QINTER12,216,28,15,11,ES_AUTOHSCROLL | ES_NUMBER | 
                    NOT WS_BORDER
    EDITTEXT        IDC_QINTER13,232,28,15,11,ES_AUTOHSCROLL | ES_NUMBER | 
                    NOT WS_BORDER
    EDITTEXT        IDC_QINTER14,248,28,15,11,ES_AUTOHSCROLL | ES_NUMBER | 
                    NOT WS_BORDER
    EDITTEXT        IDC_QINTER15,264,28,15,11,ES_AUTOHSCROLL | ES_NUMBER | 
                    NOT WS_BORDER
    EDITTEXT        IDC_QINTER16,152,40,15,11,ES_AUTOHSCROLL | ES_NUMBER | 
                    NOT WS_BORDER
    EDITTEXT        IDC_QINTER17,168,40,15,11,ES_AUTOHSCROLL | ES_NUMBER | 
                    NOT WS_BORDER
    EDITTEXT        IDC_QINTER18,184,40,15,11,ES_AUTOHSCROLL | ES_NUMBER | 
                    NOT WS_BORDER
    EDITTEXT        IDC_QINTER19,200,40,15,11,ES_AUTOHSCROLL | ES_NUMBER | 
                    NOT WS_BORDER
    EDITTEXT        IDC_QINTER20,216,40,15,11,ES_AUTOHSCROLL | ES_NUMBER | 
                    NOT WS_BORDER
    EDITTEXT        IDC_QINTER21,232,40,15,11,ES_AUTOHSCROLL | ES_NUMBER | 
                    NOT WS_BORDER
    EDITTEXT        IDC_QINTER22,248,40,15,11,ES_AUTOHSCROLL | ES_NUMBER | 
                    NOT WS_BORDER
    EDITTEXT        IDC_QINTER23,264,40,15,11,ES_AUTOHSCROLL | ES_NUMBER | 
                    NOT WS_BORDER
    EDITTEXT        IDC_QINTER24,152,52,15,11,ES_AUTOHSCROLL | ES_NUMBER | 
                    NOT WS_BORDER
    EDITTEXT        IDC_QINTER25,168,52,15,11,ES_AUTOHSCROLL | ES_NUMBER | 
                    NOT WS_BORDER
    EDITTEXT        IDC_QINTER26,184,52,15,11,ES_AUTOHSCROLL | ES_NUMBER | 
                    NOT WS_BORDER
    EDITTEXT        IDC_QINTER27,200,52,15,11,ES_AUTOHSCROLL | ES_NUMBER | 
                    NOT WS_BORDER
    EDITTEXT        IDC_QINTER28,216,52,15,11,ES_AUTOHSCROLL | ES_NUMBER | 
                    NOT WS_BORDER
    EDITTEXT        IDC_QINTER29,232,52,15,11,ES_AUTOHSCROLL | ES_NUMBER | 
                    NOT WS_BORDER
    EDITTEXT        IDC_QINTER30,248,52,15,11,ES_AUTOHSCROLL | ES_NUMBER | 
                    NOT WS_BORDER
    EDITTEXT        IDC_QINTER31,264,52,15,11,ES_AUTOHSCROLL | ES_NUMBER | 
                    NOT WS_BORDER
    EDITTEXT        IDC_QINTER32,152,64,15,11,ES_AUTOHSCROLL | ES_NUMBER | 
                    NOT WS_BORDER
    EDITTEXT        IDC_QINTER33,168,64,15,11,ES_AUTOHSCROLL | ES_NUMBER | 
                    NOT WS_BORDER
    EDITTEXT        IDC_QINTER34,184,64,15,11,ES_AUTOHSCROLL | ES_NUMBER | 
                    NOT WS_BORDER
    EDITTEXT        IDC_QINTER35,200,64,15,11,ES_AUTOHSCROLL | ES_NUMBER | 
                    NOT WS_BORDER
    EDITTEXT        IDC_QINTER36,216,64,15,11,ES_AUTOHSCROLL | ES_NUMBER | 
                    NOT WS_BORDER
    EDITTEXT        IDC_QINTER37,232,64,15,11,ES_AUTOHSCROLL | ES_NUMBER | 
                    NOT WS_BORDER
    EDITTEXT        IDC_QINTER38,248,64,15,11,ES_AUTOHSCROLL | ES_NUMBER | 
                    NOT WS_BORDER
    EDITTEXT        IDC_QINTER39,264,64,15,11,ES_AUTOHSCROLL | ES_NUMBER | 
                    NOT WS_BORDER
    EDITTEXT        IDC_QINTER40,152,76,15,11,ES_AUTOHSCROLL | ES_NUMBER | 
                    NOT WS_BORDER
    EDITTEXT        IDC_QINTER41,168,76,15,11,ES_AUTOHSCROLL | ES_NUMBER | 
                    NOT WS_BORDER
    EDITTEXT        IDC_QINTER42,184,76,15,11,ES_AUTOHSCROLL | ES_NUMBER | 
                    NOT WS_BORDER
    EDITTEXT        IDC_QINTER43,200,76,15,11,ES_AUTOHSCROLL | ES_NUMBER | 
                    NOT WS_BORDER
    EDITTEXT        IDC_QINTER44,216,76,15,11,ES_AUTOHSCROLL | ES_NUMBER | 
                    NOT WS_BORDER
    EDITTEXT        IDC_QINTER45,232,76,15,11,ES_AUTOHSCROLL | ES_NUMBER | 
                    NOT WS_BORDER
    EDITTEXT        IDC_QINTER46,248,76,15,11,ES_AUTOHSCROLL | ES_NUMBER | 
                    NOT WS_BORDER
    EDITTEXT        IDC_QINTER47,264,76,15,11,ES_AUTOHSCROLL | ES_NUMBER | 
                    NOT WS_BORDER
    EDITTEXT        IDC_QINTER48,152,88,15,11,ES_AUTOHSCROLL | ES_NUMBER | 
                    NOT WS_BORDER
    EDITTEXT        IDC_QINTER49,168,88,15,11,ES_AUTOHSCROLL | ES_NUMBER | 
                    NOT WS_BORDER
    EDITTEXT        IDC_QINTER50,184,88,15,11,ES_AUTOHSCROLL | ES_NUMBER | 
                    NOT WS_BORDER
    EDITTEXT        IDC_QINTER51,200,88,15,11,ES_AUTOHSCROLL | ES_NUMBER | 
                    NOT WS_BORDER
    EDITTEXT        IDC_QINTER52,216,88,15,11,ES_AUTOHSCROLL | ES_NUMBER | 
                    NOT WS_BORDER
    EDITTEXT        IDC_QINTER53,232,88,15,11,ES_AUTOHSCROLL | ES_NUMBER | 
                    NOT WS_BORDER
    EDITTEXT        IDC_QINTER54,248,88,15,11,ES_AUTOHSCROLL | ES_NUMBER | 
                    NOT WS_BORDER
    EDITTEXT        IDC_QINTER55,264,88,15,11,ES_AUTOHSCROLL | ES_NUMBER | 
                    NOT WS_BORDER
    EDITTEXT        IDC_QINTER56,152,100,15,11,ES_AUTOHSCROLL | ES_NUMBER | 
                    NOT WS_BORDER
    EDITTEXT        IDC_QINTER57,168,100,15,11,ES_AUTOHSCROLL | ES_NUMBER | 
                    NOT WS_BORDER
    EDITTEXT        IDC_QINTER58,184,100,15,11,ES_AUTOHSCROLL | ES_NUMBER | 
                    NOT WS_BORDER
    EDITTEXT        IDC_QINTER59,200,100,15,11,ES_AUTOHSCROLL | ES_NUMBER | 
                    NOT WS_BORDER
    EDITTEXT        IDC_QINTER60,216,100,15,11,ES_AUTOHSCROLL | ES_NUMBER | 
                    NOT WS_BORDER
    EDITTEXT        IDC_QINTER61,232,100,15,11,ES_AUTOHSCROLL | ES_NUMBER | 
                    NOT WS_BORDER
    EDITTEXT        IDC_QINTER62,248,100,15,11,ES_AUTOHSCROLL | ES_NUMBER | 
                    NOT WS_BORDER
    EDITTEXT        IDC_QINTER63,264,100,15,11,ES_AUTOHSCROLL | ES_NUMBER | 
                    NOT WS_BORDER
    GROUPBOX        "Inter matrix",IDC_STATIC,148,4,136,112
    PUSHBUTTON      "&Load matrix...",IDC_LOAD,68,128,47,13
    PUSHBUTTON      "&Save matrix...",IDC_SAVE,120,128,47,13
END

IDD_ABOUT DIALOG DISCARDABLE  0, 0, 192, 165
STYLE DS_MODALFRAME | WS_POPUP | WS_CAPTION | WS_SYSMENU
CAPTION "XviD MPEG4 Video Codec"
FONT 8, "MS Shell Dlg"
BEGIN
    CTEXT           "GPL'ed MPEG4 video codec demonstration.\nFor educational purposes only.",
                    IDC_STATIC,12,112,168,20
    CTEXT           "WEBSITE",IDC_WEBSITE,60,92,72,8,SS_NOTIFY | 
                    SS_CENTERIMAGE
    CTEXT           "BUILD",IDC_BUILD,8,28,176,8,SS_CENTERIMAGE
    CONTROL         "IDB_LOGO",IDC_STATIC,"Static",SS_BITMAP,24,56,15,13
    CTEXT           "XviD MPEG4 video codec",IDC_STATIC,8,16,176,12
    GROUPBOX        "About",IDC_STATIC,4,4,184,132
    DEFPUSHBUTTON   "OK",IDOK,102,144,80,14
    CTEXT           "CORE",IDC_CORE,8,40,176,8,SS_CENTERIMAGE
    CTEXT           "( SPECIAL BUILD )",IDC_SPECIAL_BUILD,5,102,181,8
    PUSHBUTTON      "View License...",IDC_LICENSE,10,144,80,14
END

IDD_RC_CBR DIALOG DISCARDABLE  0, 0, 212, 212
STYLE DS_MODALFRAME | WS_POPUP | WS_CAPTION | WS_SYSMENU
CAPTION "CBR"
FONT 8, "MS Shell Dlg"
BEGIN
    EDITTEXT        IDC_CBR_REACTIONDELAY,108,12,76,12,ES_AUTOHSCROLL | 
                    ES_NUMBER
    LTEXT           "Reaction Delay Factor:",IDC_STATIC,8,12,80,12,
                    SS_CENTERIMAGE
    EDITTEXT        IDC_CBR_AVERAGINGPERIOD,108,28,76,12,ES_AUTOHSCROLL | 
                    ES_NUMBER
    LTEXT           "Averaging period:",IDC_STATIC,8,28,80,12,SS_CENTERIMAGE
    EDITTEXT        IDC_CBR_BUFFER,108,44,76,12,ES_AUTOHSCROLL | ES_NUMBER
    LTEXT           "Smoother:",IDC_STATIC,8,44,80,12,SS_CENTERIMAGE
END

IDD_PROFILE DIALOG DISCARDABLE  0, 0, 212, 212
STYLE DS_MODALFRAME | WS_POPUP | WS_CAPTION | WS_SYSMENU
CAPTION "Profile"
FONT 8, "MS Shell Dlg"
BEGIN
    GROUPBOX        "",IDC_STATIC,8,120,196,86
    LTEXT           "Profile @ Level:",IDC_STATIC,8,10,52,8
    COMBOBOX        IDC_PROFILE_PROFILE,84,8,116,200,CBS_DROPDOWNLIST | 
                    WS_VSCROLL | WS_TABSTOP
    COMBOBOX        IDC_QUANTTYPE,111,32,76,76,CBS_DROPDOWNLIST | WS_VSCROLL | 
                    WS_TABSTOP
    PUSHBUTTON      "Edit Matrix...",IDC_QUANTMATRIX,111,47,76,12
    CONTROL         "Adaptive Quantization",IDC_LUMMASK,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,16,56,85,10
    CONTROL         "Interlaced Encoding",IDC_INTERLACING,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,16,72,79,8
    CONTROL         "Quarter Pixel",IDC_QPEL,"Button",BS_AUTOCHECKBOX | 
                    WS_TABSTOP,16,88,100,8
    CONTROL         "Global Motion Compensation",IDC_GMC,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,16,104,103,8
    CONTROL         "B-VOPs",IDC_BVOP,"Button",BS_AUTOCHECKBOX | WS_TABSTOP,
                    16,120,40,8
    LTEXT           "Max consecutive BVOPs:",IDC_MAXBFRAMES_S,16,136,96,8
    EDITTEXT        IDC_MAXBFRAMES,120,133,76,12,ES_AUTOHSCROLL | ES_NUMBER
    EDITTEXT        IDC_BQUANTRATIO,120,147,76,12,ES_AUTOHSCROLL
    EDITTEXT        IDC_BQUANTOFFSET,120,161,76,12,ES_AUTOHSCROLL
    CONTROL         "Packed bitstream",IDC_PACKED,"Button",BS_AUTOCHECKBOX | 
                    WS_TABSTOP,16,182,71,10
    LTEXT           "Quantizer ratio:",IDC_BQUANTRATIO_S,16,150,89,8
    LTEXT           "Quantizer offset:",IDC_BQUANTOFFSET_S,16,164,52,8
    LTEXT           "Quantization type:",IDC_QUANTTYPE_S,16,36,85,8
    CONTROL         "Top field first",IDC_TFF,"Button",BS_AUTOCHECKBOX | 
                    WS_TABSTOP,100,72,56,10
END

IDD_ZONE DIALOG DISCARDABLE  0, 0, 212, 194
STYLE DS_MODALFRAME | WS_POPUP | WS_CAPTION | WS_SYSMENU
CAPTION "Zone"
FONT 8, "MS Shell Dlg"
BEGIN
    EDITTEXT        IDC_ZONE_FRAME,104,6,80,12,ES_AUTOHSCROLL | ES_NUMBER
    PUSHBUTTON      "<-",IDC_ZONE_FETCH,188,6,16,12
    CONTROL         "Weight:",IDC_ZONE_MODE_WEIGHT,"Button",
                    BS_AUTORADIOBUTTON | WS_GROUP,16,40,41,10
    EDITTEXT        IDC_ZONE_WEIGHT,104,38,80,12,ES_AUTOHSCROLL
    CONTROL         "Quantizer:",IDC_ZONE_MODE_QUANT,"Button",
                    BS_AUTORADIOBUTTON,16,56,48,10
    EDITTEXT        IDC_ZONE_QUANT,104,54,80,12,ES_AUTOHSCROLL
    CONTROL         "Slider1",IDC_ZONE_SLIDER,"msctls_trackbar32",TBS_BOTH | 
                    TBS_NOTICKS | WS_TABSTOP,12,82,188,14
    CONTROL         "Begin with keyframe",IDC_ZONE_FORCEIVOP,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,16,124,92,8
    CONTROL         "Greyscale encoding",IDC_ZONE_GREYSCALE,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,16,138,79,10
    CONTROL         "Chroma optimizer enabled",IDC_ZONE_CHROMAOPT,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,16,152,97,10
    LTEXT           "BVOP sensitivity:",IDC_ZONE_BVOPTHRESHOLD_S,16,170,76,8
    EDITTEXT        IDC_ZONE_BVOPTHRESHOLD,100,168,84,12,ES_AUTOHSCROLL
    LTEXT           "Start frame #:",IDC_STATIC,8,6,52,10
    GROUPBOX        "Rate control",IDC_STATIC,7,22,198,90
    RTEXT           "X",IDC_ZONE_MAX,140,74,54,8
    LTEXT           "X",IDC_ZONE_MIN,18,74,54,8
    GROUPBOX        "Static",IDC_STATIC,7,107,198,80
    CONTROL         "Cartoon Mode",IDC_CARTOON,"Button",BS_AUTOCHECKBOX | 
                    WS_TABSTOP,121,124,61,10
END

IDD_LEVEL DIALOG DISCARDABLE  0, 0, 212, 215
STYLE DS_MODALFRAME | WS_POPUP | WS_CAPTION | WS_SYSMENU
CAPTION "Level"
FONT 8, "MS Shell Dlg"
BEGIN
    COMBOBOX        IDC_LEVEL_PROFILE,84,8,116,200,CBS_DROPDOWNLIST | 
                    WS_VSCROLL | WS_TABSTOP
    EDITTEXT        IDC_LEVEL_WIDTH,88,43,28,12,ES_AUTOHSCROLL | ES_READONLY
    EDITTEXT        IDC_LEVEL_HEIGHT,128,43,28,12,ES_AUTOHSCROLL | 
                    ES_READONLY
    EDITTEXT        IDC_LEVEL_FPS,168,43,28,12,ES_AUTOHSCROLL | ES_READONLY
    EDITTEXT        IDC_LEVEL_VMV,156,60,40,12,ES_RIGHT | ES_AUTOHSCROLL | 
                    ES_READONLY
    EDITTEXT        IDC_LEVEL_VCV,156,77,40,12,ES_RIGHT | ES_AUTOHSCROLL | 
                    ES_READONLY
    EDITTEXT        IDC_LEVEL_VBV,156,115,40,12,ES_RIGHT | ES_AUTOHSCROLL | 
                    ES_READONLY
    EDITTEXT        IDC_LEVEL_BITRATE,156,131,40,12,ES_RIGHT | 
                    ES_AUTOHSCROLL | ES_READONLY
    LTEXT           "Profile @ Level:",IDC_STATIC,8,10,52,8
    GROUPBOX        "Level - XviD will not force you to respect these",
                    IDC_LEVEL_LEVEL_G,8,28,196,66
    LTEXT           "Suggested:",IDC_LEVEL_DIM_S,16,45,48,8
    LTEXT           "x",IDC_STATIC,120,45,8,8
    LTEXT           "Max bitrate (kbps)",IDC_LEVEL_BITRATE_S,16,133,108,8
    LTEXT           "Max buffer size (bits):",IDC_LEVEL_VBV_S,16,117,108,8
    LTEXT           "Max processing rate (mbs/sec)",IDC_LEVEL_VCV_S,16,79,
                    108,8
    LTEXT           "Max frame size (macroblocks):",IDC_LEVEL_VMV_S,16,63,
                    108,8
    LTEXT           "x",IDC_STATIC,160,45,8,8
    GROUPBOX        "Video Buffer Verifier - used in Two-Pass mode",
                    IDC_LEVEL_VBV_G,8,100,196,66
    EDITTEXT        IDC_LEVEL_PEAKRATE,156,148,40,12,ES_RIGHT | 
                    ES_AUTOHSCROLL | ES_READONLY
    LTEXT           "Max bits over any one second interval:",
                    IDC_LEVEL_PEAKRATE_S,16,150,128,8
END

IDD_DEC DIALOG DISCARDABLE  0, 0, 212, 212
STYLE DS_MODALFRAME | WS_POPUP | WS_CAPTION | WS_SYSMENU
CAPTION "Decoder"
FONT 8, "MS Shell Dlg"
BEGIN
    CONTROL         "Deblocking Y",IDC_DEC_DY,"Button",BS_AUTOCHECKBOX | 
                    WS_TABSTOP,15,64,63,13
    CONTROL         "Deblocking UV",IDC_DEC_DUV,"Button",BS_AUTOCHECKBOX | 
                    WS_TABSTOP,15,80,61,13
    CONTROL         "Deringing Y",IDC_DEC_DRY,"Button",BS_AUTOCHECKBOX | 
                    WS_TABSTOP,88,64,60,13
    CONTROL         "Film Effect",IDC_DEC_FE,"Button",BS_AUTOCHECKBOX | 
                    WS_TABSTOP,154,64,45,13
    GROUPBOX        "Brightness",IDC_STATIC,5,6,202,41
    CONTROL         "Slider1",IDC_DEC_BRIGHTNESS,"msctls_trackbar32",
                    TBS_AUTOTICKS | TBS_BOTH | WS_TABSTOP,18,18,181,24
    GROUPBOX        "Postprocessing",IDC_STATIC,5,52,202,46
    CONTROL         "Deringing UV",IDC_DEC_DRUV,"Button",BS_AUTOCHECKBOX | 
                    WS_TABSTOP,88,80,60,13
END

IDD_STATUS DIALOGEX 0, 0, 325, 220
STYLE DS_MODALFRAME | WS_MINIMIZEBOX | WS_POPUP | WS_CAPTION | WS_SYSMENU
CAPTION "XviD Status"
FONT 8, "MS Shell Dlg", 0, 0, 0x1
BEGIN
    CONTROL         "",IDC_STATUS_GRAPH,"Static",SS_OWNERDRAW | SS_NOTIFY,4,
                    4,187,108
    LTEXT           "I-VOP",IDC_STATIC,26,139,22,8
    LTEXT           "B-VOP",IDC_STATIC,26,167,24,8
    LTEXT           "P-VOP",IDC_STATIC,26,153,23,8
    LTEXT           "Total",IDC_STATIC,26,181,21,8
    EDITTEXT        IDC_STATUS_IQ_MIN,103,137,17,12,ES_RIGHT | 
                    ES_AUTOHSCROLL | ES_READONLY | ES_NUMBER | NOT WS_BORDER | 
                    NOT WS_TABSTOP,WS_EX_STATICEDGE
    EDITTEXT        IDC_STATUS_IQ_MAX,123,137,17,12,ES_RIGHT | 
                    ES_AUTOHSCROLL | ES_READONLY | ES_NUMBER | NOT WS_BORDER | 
                    NOT WS_TABSTOP,WS_EX_STATICEDGE
    EDITTEXT        IDC_STATUS_PQ_MIN,103,152,17,12,ES_RIGHT | 
                    ES_AUTOHSCROLL | ES_READONLY | ES_NUMBER | NOT WS_BORDER | 
                    NOT WS_TABSTOP,WS_EX_STATICEDGE
    EDITTEXT        IDC_STATUS_PQ_MAX,123,152,17,12,ES_RIGHT | 
                    ES_AUTOHSCROLL | ES_READONLY | ES_NUMBER | NOT WS_BORDER | 
                    NOT WS_TABSTOP,WS_EX_STATICEDGE
    EDITTEXT        IDC_STATUS_BQ_MIN,103,166,17,12,ES_RIGHT | 
                    ES_AUTOHSCROLL | ES_READONLY | ES_NUMBER | NOT WS_BORDER | 
                    NOT WS_TABSTOP,WS_EX_STATICEDGE
    EDITTEXT        IDC_STATUS_BQ_MAX,123,166,17,12,ES_RIGHT | 
                    ES_AUTOHSCROLL | ES_READONLY | ES_NUMBER | NOT WS_BORDER | 
                    NOT WS_TABSTOP,WS_EX_STATICEDGE
    EDITTEXT        IDC_STATUS_Q_MIN,103,180,17,12,ES_RIGHT | ES_AUTOHSCROLL | 
                    ES_READONLY | ES_NUMBER | NOT WS_BORDER | NOT WS_TABSTOP,
                    WS_EX_STATICEDGE
    EDITTEXT        IDC_STATUS_Q_MAX,123,180,17,12,ES_RIGHT | ES_AUTOHSCROLL | 
                    ES_READONLY | ES_NUMBER | NOT WS_BORDER | NOT WS_TABSTOP,
                    WS_EX_STATICEDGE
    CTEXT           "Min",IDC_STATIC,104,126,14,8
    CTEXT           "Max",IDC_STATIC,123,126,16,8
    EDITTEXT        IDC_STATUS_IL_MIN,174,137,29,12,ES_RIGHT | 
                    ES_AUTOHSCROLL | ES_READONLY | ES_NUMBER | NOT WS_BORDER | 
                    NOT WS_TABSTOP,WS_EX_STATICEDGE
    EDITTEXT        IDC_STATUS_IL_MAX,206,137,28,12,ES_RIGHT | 
                    ES_AUTOHSCROLL | ES_READONLY | ES_NUMBER | NOT WS_BORDER | 
                    NOT WS_TABSTOP,WS_EX_STATICEDGE
    EDITTEXT        IDC_STATUS_PL_MIN,174,152,29,12,ES_RIGHT | 
                    ES_AUTOHSCROLL | ES_READONLY | ES_NUMBER | NOT WS_BORDER | 
                    NOT WS_TABSTOP,WS_EX_STATICEDGE
    EDITTEXT        IDC_STATUS_PL_MAX,206,152,28,12,ES_RIGHT | 
                    ES_AUTOHSCROLL | ES_READONLY | ES_NUMBER | NOT WS_BORDER | 
                    NOT WS_TABSTOP,WS_EX_STATICEDGE
    EDITTEXT        IDC_STATUS_BL_MIN,174,166,29,12,ES_RIGHT | 
                    ES_AUTOHSCROLL | ES_READONLY | ES_NUMBER | NOT WS_BORDER | 
                    NOT WS_TABSTOP,WS_EX_STATICEDGE
    EDITTEXT        IDC_STATUS_BL_MAX,206,166,28,12,ES_RIGHT | 
                    ES_AUTOHSCROLL | ES_READONLY | ES_NUMBER | NOT WS_BORDER | 
                    NOT WS_TABSTOP,WS_EX_STATICEDGE
    EDITTEXT        IDC_STATUS_L_MIN,174,180,29,12,ES_RIGHT | ES_AUTOHSCROLL | 
                    ES_READONLY | ES_NUMBER | NOT WS_BORDER | NOT WS_TABSTOP,
                    WS_EX_STATICEDGE
    EDITTEXT        IDC_STATUS_L_MAX,206,180,28,12,ES_RIGHT | ES_AUTOHSCROLL | 
                    ES_READONLY | ES_NUMBER | NOT WS_BORDER | NOT WS_TABSTOP,
                    WS_EX_STATICEDGE
    CTEXT           "Min",IDC_STATIC,177,125,24,8
    CTEXT           "Max",IDC_STATIC,205,125,29,8
    EDITTEXT        IDC_STATUS_IL_TOT,270,137,29,12,ES_RIGHT | 
                    ES_AUTOHSCROLL | ES_READONLY | ES_NUMBER | NOT WS_BORDER | 
                    NOT WS_TABSTOP,WS_EX_STATICEDGE
    EDITTEXT        IDC_STATUS_PL_TOT,270,152,29,12,ES_RIGHT | 
                    ES_AUTOHSCROLL | ES_READONLY | ES_NUMBER | NOT WS_BORDER | 
                    NOT WS_TABSTOP,WS_EX_STATICEDGE
    EDITTEXT        IDC_STATUS_BL_TOT,270,166,29,12,ES_RIGHT | 
                    ES_AUTOHSCROLL | ES_READONLY | ES_NUMBER | NOT WS_BORDER | 
                    NOT WS_TABSTOP,WS_EX_STATICEDGE
    EDITTEXT        IDC_STATUS_L_TOT,270,180,29,12,ES_RIGHT | ES_AUTOHSCROLL | 
                    ES_READONLY | ES_NUMBER | NOT WS_BORDER | NOT WS_TABSTOP,
                    WS_EX_STATICEDGE
    CTEXT           "Total (k)",IDC_STATIC,271,125,28,8
    CONTROL         "Auto-close window",IDC_STATUS_DESTROY,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,13,202,100,10
    EDITTEXT        IDC_STATUS_KBPS,267,199,36,12,ES_RIGHT | ES_AUTOHSCROLL | 
                    ES_READONLY | ES_NUMBER | NOT WS_BORDER | NOT WS_TABSTOP,
                    WS_EX_STATICEDGE
    EDITTEXT        IDC_STATUS_IL_AVG,237,137,29,12,ES_RIGHT | 
                    ES_AUTOHSCROLL | ES_READONLY | ES_NUMBER | NOT WS_BORDER | 
                    NOT WS_TABSTOP,WS_EX_STATICEDGE
    EDITTEXT        IDC_STATUS_PL_AVG,237,152,29,12,ES_RIGHT | 
                    ES_AUTOHSCROLL | ES_READONLY | ES_NUMBER | NOT WS_BORDER | 
                    NOT WS_TABSTOP,WS_EX_STATICEDGE
    EDITTEXT        IDC_STATUS_BL_AVG,237,166,29,12,ES_RIGHT | 
                    ES_AUTOHSCROLL | ES_READONLY | ES_NUMBER | NOT WS_BORDER | 
                    NOT WS_TABSTOP,WS_EX_STATICEDGE
    EDITTEXT        IDC_STATUS_L_AVG,237,180,29,12,ES_RIGHT | ES_AUTOHSCROLL | 
                    ES_READONLY | ES_NUMBER | NOT WS_BORDER | NOT WS_TABSTOP,
                    WS_EX_STATICEDGE
    CTEXT           "Average",IDC_STATIC,237,125,30,8
    EDITTEXT        IDC_STATUS_I_NUM,57,138,32,12,ES_RIGHT | ES_AUTOHSCROLL | 
                    ES_READONLY | ES_NUMBER | NOT WS_BORDER | NOT WS_TABSTOP,
                    WS_EX_STATICEDGE
    EDITTEXT        IDC_STATUS_P_NUM,57,152,32,12,ES_RIGHT | ES_AUTOHSCROLL | 
                    ES_READONLY | ES_NUMBER | NOT WS_BORDER | NOT WS_TABSTOP,
                    WS_EX_STATICEDGE
    EDITTEXT        IDC_STATUS_B_NUM,57,166,32,12,ES_RIGHT | ES_AUTOHSCROLL | 
                    ES_READONLY | ES_NUMBER | NOT WS_BORDER | NOT WS_TABSTOP,
                    WS_EX_STATICEDGE
    EDITTEXT        IDC_STATUS_NUM,57,180,32,12,ES_RIGHT | ES_AUTOHSCROLL | 
                    ES_READONLY | ES_NUMBER | NOT WS_BORDER | NOT WS_TABSTOP,
                    WS_EX_STATICEDGE
    GROUPBOX        "Frame size (bytes)",IDC_STATIC,170,117,133,80,BS_CENTER
    GROUPBOX        "Quant",IDC_STATIC,99,117,65,80,BS_CENTER
    GROUPBOX        "Frames",IDC_STATIC,53,117,40,80,BS_CENTER
    LTEXT           "Avg bitrate (kbps):",IDC_STATIC,200,201,59,8
    LISTBOX         IDC_DEBUGOUTPUT,201,13,113,100,LBS_NOINTEGRALHEIGHT | 
                    WS_TABSTOP
    CONTROL         "Show me the internals!",IDC_SHOWINTERNALS,"Button",
                    BS_AUTOCHECKBOX | BS_NOTIFY | WS_TABSTOP,201,1,87,10
    EDITTEXT        IDC_STATUS_IQ_AVG,143,137,17,12,ES_RIGHT | 
                    ES_AUTOHSCROLL | ES_READONLY | ES_NUMBER | NOT WS_BORDER | 
                    NOT WS_TABSTOP,WS_EX_STATICEDGE
    EDITTEXT        IDC_STATUS_PQ_AVG,143,152,17,12,ES_RIGHT | 
                    ES_AUTOHSCROLL | ES_READONLY | ES_NUMBER | NOT WS_BORDER | 
                    NOT WS_TABSTOP,WS_EX_STATICEDGE
    EDITTEXT        IDC_STATUS_BQ_AVG,143,166,17,12,ES_RIGHT | 
                    ES_AUTOHSCROLL | ES_READONLY | ES_NUMBER | NOT WS_BORDER | 
                    NOT WS_TABSTOP,WS_EX_STATICEDGE
    EDITTEXT        IDC_STATUS_Q_AVG,143,180,17,12,ES_RIGHT | ES_AUTOHSCROLL | 
                    ES_READONLY | ES_NUMBER | NOT WS_BORDER | NOT WS_TABSTOP,
                    WS_EX_STATICEDGE
    CTEXT           "Avg",IDC_STATIC,143,126,16,8
END

IDD_AR DIALOGEX 0, 0, 211, 215
STYLE DS_MODALFRAME | WS_POPUP | WS_CAPTION | WS_SYSMENU
CAPTION "Aspect Ratio"
FONT 8, "MS Shell Dlg", 0, 0, 0x1
BEGIN
    CONTROL         "Pixel Aspect Ratio",IDC_PAR,"Button",BS_AUTORADIOBUTTON,
                    13,7,73,10,WS_EX_TRANSPARENT
    COMBOBOX        IDC_ASPECT_RATIO,25,33,111,55,CBS_DROPDOWNLIST | 
                    WS_VSCROLL | WS_TABSTOP
    EDITTEXT        IDC_PARX,39,58,36,12,ES_NUMBER
    EDITTEXT        IDC_PARY,100,58,36,12,ES_NUMBER
    PUSHBUTTON      "Default",IDC_AR_DEFAULT,156,21,37,11
    CONTROL         "Display Aspect Ratio",IDC_AR,"Button",
                    BS_AUTORADIOBUTTON,13,81,81,10,WS_EX_TRANSPARENT
    EDITTEXT        IDC_ARX,39,110,36,12,ES_NUMBER
    EDITTEXT        IDC_ARY,100,110,36,12,ES_NUMBER
    PUSHBUTTON      "4:3",IDC_AR_4_3,156,90,37,11
    PUSHBUTTON      "16:9",IDC_AR_16_9,156,104,37,11
    PUSHBUTTON      "2,35:1",IDC_AR_235_100,156,118,37,11
    LTEXT           "Select the shape of a pixel...",IDC_STATIC,25,18,104,10
    LTEXT           "Select the shape of the image...",IDC_STATIC,25,93,110,
                    11
    GROUPBOX        "",IDC_STATIC,7,7,196,127,BS_CENTER
    GROUPBOX        "",IDC_STATIC,7,81,196,52,BS_CENTER
    CTEXT           "X :",IDC_STATIC,25,60,10,10
    LTEXT           "Aspect Ratio is written to mpeg-4 bitstream, but unfortunately is likely to be ignored if video stream is encapsulated in a general-purpose container (like .avi, .ogm., .mkv).",
                    IDC_STATIC,7,138,197,27
    LTEXT           "Therefore, be aware that using different aspect ratio than default might be ignored by some players, especially when decoded on Windows.\n\nUse at your own risk.",
                    IDC_STATIC,7,168,197,41
    CTEXT           "Y :",IDC_STATIC,86,60,10,10
    CTEXT           "Y :",IDC_STATIC,86,112,10,10
    CTEXT           "X :",IDC_STATIC,25,112,10,10
    GROUPBOX        "Quick Setting",IDC_STATIC,148,7,55,127,0,
                    WS_EX_TRANSPARENT
END

IDD_BITRATE DIALOG DISCARDABLE  0, 0, 212, 212
STYLE DS_MODALFRAME | WS_POPUP | WS_CAPTION | WS_SYSMENU
CAPTION "Bitrate Calculator"
FONT 8, "MS Shell Dlg"
BEGIN
    COMBOBOX        IDC_BITRATE_TSIZE,95,5,75,64,CBS_DROPDOWN | WS_VSCROLL | 
                    WS_TABSTOP
    EDITTEXT        IDC_BITRATE_SSIZE,95,20,75,12,ES_AUTOHSCROLL
    PUSHBUTTON      "...",IDC_BITRATE_SSELECT,178,21,20,12
    COMBOBOX        IDC_BITRATE_CFORMAT,95,45,75,64,CBS_DROPDOWNLIST | 
                    WS_VSCROLL | WS_TABSTOP
    EDITTEXT        IDC_BITRATE_COVERHEAD,95,60,75,12,ES_AUTOHSCROLL | 
                    ES_READONLY
    EDITTEXT        IDC_BITRATE_HOURS,15,100,29,11,ES_AUTOHSCROLL
    EDITTEXT        IDC_BITRATE_MINUTES,52,100,29,11,ES_AUTOHSCROLL
    EDITTEXT        IDC_BITRATE_SECONDS,88,100,29,11,ES_AUTOHSCROLL
    COMBOBOX        IDC_BITRATE_FPS,124,100,71,100,CBS_DROPDOWNLIST | 
                    WS_VSCROLL | WS_TABSTOP
    EDITTEXT        IDC_BITRATE_VSIZE,95,120,75,12,ES_AUTOHSCROLL | 
                    ES_READONLY
    EDITTEXT        IDC_BITRATE_VRATE,95,135,75,12,ES_AUTOHSCROLL | 
                    ES_READONLY
    COMBOBOX        IDC_BITRATE_AFORMAT,95,158,75,64,CBS_DROPDOWNLIST | 
                    WS_VSCROLL | WS_TABSTOP
    COMBOBOX        IDC_BITRATE_ARATE,95,173,75,64,CBS_DROPDOWN | WS_VSCROLL | 
                    WS_TABSTOP
    EDITTEXT        IDC_BITRATE_ASIZE,95,188,75,12,ES_AUTOHSCROLL
    PUSHBUTTON      "...",IDC_BITRATE_ASELECT,178,188,20,12
    CONTROL         "Avg. bitrate (kbps):",IDC_BITRATE_AMODE_RATE,"Button",
                    BS_AUTORADIOBUTTON,15,175,75,10
    GROUPBOX        "Video",IDC_STATIC,5,75,200,132
    GROUPBOX        "Audio",IDC_STATIC,5,148,200,59
    CONTROL         "Size (kbytes):",IDC_BITRATE_AMODE_SIZE,"Button",
                    BS_AUTORADIOBUTTON,15,190,58,10
    LTEXT           "Target size (kbytes):",IDC_STATIC,15,6,64,12,
                    SS_CENTERIMAGE
    LTEXT           "Format:",IDC_STATIC,15,46,24,13,SS_CENTERIMAGE
    LTEXT           "Format:",IDC_STATIC,15,160,24,8,SS_CENTERIMAGE
    LTEXT           "Size (kbytes):",IDC_STATIC,15,121,43,8,SS_CENTERIMAGE
    GROUPBOX        "Container:",IDC_STATIC,5,35,200,172
    LTEXT           "Overhead (kbytes):",IDC_STATIC,15,61,61,10,
                    SS_CENTERIMAGE
    LTEXT           "Average bitrate (kbps):",IDC_STATIC,15,136,72,8,
                    SS_CENTERIMAGE
    CTEXT           "hours",IDC_STATIC,15,90,30,8
    CTEXT           "minutes",IDC_STATIC,52,90,30,8
    CTEXT           "seconds",IDC_STATIC,88,90,30,8
    CTEXT           "frames per second",IDC_STATIC,126,90,70,8
    LTEXT           "Subtitles (kbytes):",IDC_STATIC,15,21,60,10,
                    SS_CENTERIMAGE
END

IDD_LICENSE DIALOG DISCARDABLE  0, 0, 430, 234
STYLE DS_MODALFRAME | WS_POPUP | WS_CAPTION | WS_SYSMENU
CAPTION "GNU General Public License"
FONT 8, "MS Sans Serif"
BEGIN
    PUSHBUTTON      "OK",IDOK,172,218,84,14
    EDITTEXT        IDC_LICENSE_TEXT,2,2,426,212,ES_MULTILINE | 
                    ES_AUTOHSCROLL | ES_READONLY | WS_VSCROLL
END

IDD_COMMON DIALOG DISCARDABLE  0, 0, 212, 212
STYLE DS_MODALFRAME | WS_POPUP | WS_CAPTION | WS_SYSMENU
CAPTION "Common"
FONT 8, "MS Shell Dlg"
BEGIN
    CONTROL         "Automatically detect optimizations",IDC_CPU_AUTO,"Button",
                    BS_AUTORADIOBUTTON | WS_GROUP,16,20,121,10
    CONTROL         "Force optimizations",IDC_CPU_FORCE,"Button",
                    BS_AUTORADIOBUTTON,16,33,76,10
    CONTROL         "MMX",IDC_CPU_MMX,"Button",BS_AUTOCHECKBOX | WS_TABSTOP,
                    28,44,33,10
    CONTROL         "Integer SSE",IDC_CPU_MMXEXT,"Button",BS_AUTOCHECKBOX | 
                    WS_TABSTOP,28,57,54,10
    CONTROL         "SSE",IDC_CPU_SSE,"Button",BS_AUTOCHECKBOX | WS_TABSTOP,
                    28,68,30,10
    CONTROL         "SSE2",IDC_CPU_SSE2,"Button",BS_AUTOCHECKBOX | 
                    WS_TABSTOP,28,81,34,10
    CONTROL         "3DNow!",IDC_CPU_3DNOW,"Button",BS_AUTOCHECKBOX | 
                    WS_TABSTOP,28,92,42,10
    CONTROL         "3DNow! 2",IDC_CPU_3DNOWEXT,"Button",BS_AUTOCHECKBOX | 
                    WS_TABSTOP,28,105,48,10
    EDITTEXT        IDC_DEBUG,127,133,76,12,ES_AUTOHSCROLL
    GROUPBOX        "Performance optimizations",-1,8,4,196,120
    LTEXT           "OutputDebugString debug level:",-1,7,135,104,12
END


/////////////////////////////////////////////////////////////////////////////
//
// Bitmap
//

IDB_LOGO                BITMAP  DISCARDABLE     "XviD_logo.bmp"

/////////////////////////////////////////////////////////////////////////////
//
// Icon
//

// Icon with lowest ID value placed first to ensure application icon
// remains consistent on all systems.
IDI_ICON                ICON    DISCARDABLE     "xvid.ico"

/////////////////////////////////////////////////////////////////////////////
//
// TEXT
//

IDR_GPL                 TEXT    DISCARDABLE     "../LICENSE"

#ifdef APSTUDIO_INVOKED
/////////////////////////////////////////////////////////////////////////////
//
// TEXTINCLUDE
//

1 TEXTINCLUDE DISCARDABLE 
BEGIN
    "resource.h\0"
END

2 TEXTINCLUDE DISCARDABLE 
BEGIN
    "#include <windows.h>\r\n"
    "#ifndef IDC_STATIC\r\n"
    "#define IDC_STATIC (-1)\r\n"
    "#endif\r\n"
    "\0"
END

3 TEXTINCLUDE DISCARDABLE 
BEGIN
    "\r\n"
    "\0"
END

#endif    // APSTUDIO_INVOKED


/////////////////////////////////////////////////////////////////////////////
//
// String Table
//

STRINGTABLE DISCARDABLE 
BEGIN
    IDC_MODE                """Single Pass""encodes to your specified bitrate or quality.\n""Two-pass 1st pass"" gathers statistics for the 2nd pass.\n""Two-pass 2nd pass"" uses 1st pass' statistics to encode at target file size or bitrate."
END

STRINGTABLE DISCARDABLE 
BEGIN
    IDC_INTERLACING         "Encodes frames as interlaced - only use if your source contains interlacing artifacts (i.e. fields instead of progressive frames)"
    IDC_OVERDEG             "How much of the overflow the codec can eat into during oversized sections\nLarger values will prevent oversized files better, but will also spoil quantizer distribution more."
    IDC_MAXBFRAMES          "Maximum number of sequential B-frames."
    IDC_BQUANTRATIO         "Ratio used to calculate the b-frame quantizer.\n\nBVOP quant = (AVG(past VOP quant, future VOP quant) * quant ratio + quant offset)"
    IDC_OVERIMP             "How much of the overflow the codec can eat into during undersized sections.\nLarger values will prevent undersized files better, but will also spoil quantizer distribution more."
    IDC_CBR_REACTIONDELAY   "Determines how slowly it will adjust the current encoding quality based upon scene intensity - this has the strongest influence on quality"
    IDC_CBR_AVERAGINGPERIOD "Determines how slowly it adapts to the current adjusted quality"
    IDC_CBR_BUFFER          "Provides said number of frames worth of buffer between the adjusted encoding quality and lowest possible quality"
    IDC_PACKED              "When enabled the P-frames and B-frames are packed together in the one bitstreams. This permits decoding without delay.\neg. [I] [PB] [B] [empty] [PB] [B] [empty] [P]\n\nPacked bitstreams were introduced in DivX 5.01."
END

STRINGTABLE DISCARDABLE 
BEGIN
    IDC_FRAMEDROP           "Frame dropping ratio. 0 = no frame dropping .. 100 = drop all frames."
    IDC_KFREDUCTION         "Reduction of bitrate for the first consecutive i-frames. The last i-frame will get treated normally."
    IDC_GMC                 "Use Global Motion Compensation."
END

STRINGTABLE DISCARDABLE 
BEGIN
    IDC_PROFILE             "Restrict the usage of MPEG-4 tools and limit the encoded bitrate such that bitstreams are compatible with hardware decoders."
    IDC_MOTION              "Higher settings give higher-quality results, at the cost of slower encoding."
    IDC_QUANTTYPE           "H.263 smooths the image whereas MPEG (slightly slower) sharpens.\nCustom lets you define your own matrix."
    IDC_FOURCC              "Choose what you would like the avi to identify itself as"
    IDC_MAXKEY              "Maximum number of frames allowed between I-frames"
    IDC_LUMMASK             "Turns on Lumi masking - applies more compression to dark/light areas that the eye can't notice easily"
    IDC_MINIQUANT           "Minimum quantizer allowed for I-frames. Only functional in 2-pass second pass."
    IDC_MAXIQUANT           "Maximum quantizer allowed for I-frames. Only functional in 2-pass second pass."
    IDC_MINPQUANT           "Minimum quantizer allowed for P-frames."
    IDC_MAXPQUANT           "Maximum quantizer allowed for P-frames."
    IDC_MINBQUANT           "Minimum quantizer allowed for B-frames, BEFORE ratio/offset scalling"
    IDC_MAXBQUANT           "Maximum quantizer allowed for B-frames, BEFORE ratio/offset scalling"
    IDC_QUANTMATRIX         "Define your own MPEG quantization matrices.  Quantization type must be set to ""Custom"" to affect encoding."
    IDC_KFBOOST             "A value of 20 will give 20% more bits to every I-frame"
    IDC_MINKEY              "If keyframes are close together, it might be useful to decrease the bitrate of all keyframes except the last one.\nUse this to define how close must keyframes be to be reduced."
    IDC_DISCARD1PASS        "Check this if you would like to skip the storage of the 1st pass output.  It is often very large."
END

STRINGTABLE DISCARDABLE 
BEGIN
    IDC_CURVECOMPH          "The higher this value, the more bits get taken from frames larger than the average size, and redistributed to others"
    IDC_CURVECOMPL          "The higher this value, the more bits get assigned to frames below the average frame size"
    IDC_STATS1              "Location for 1st pass stats file to be saved to"
    IDC_STATS2              "Location for 2nd pass curve stats to be loaded from"
END

STRINGTABLE DISCARDABLE 
BEGIN
    IDC_CPU_AUTO            "Enable XviD's internal CPU detection"
    IDC_CPU_FORCE           "Override XviD's internal CPU detection (not recommended)"
END

STRINGTABLE DISCARDABLE 
BEGIN
    IDC_LOAD                "Load a pair of custom intra/inter matrices"
    IDC_SAVE                "Save the current intra/inter matrices to a file"
END

STRINGTABLE DISCARDABLE 
BEGIN
    IDC_QPEL                "Use Quarter PixEL resolution for encoding for a more precise motion compensation"
    IDC_CHROMAME            "Use chroma information to detect motion (slower)."
    IDC_BQUANTOFFSET        "B-frame quantizer offset from last p-frame quantizer; refer to b-frame quant ratio (above)"
    IDC_VHQ                 "VHQ enables an additional search process to increase quality (much slower)."
END

STRINGTABLE DISCARDABLE 
BEGIN
    IDC_ZONE_CHROMAOPT      "Interpolates colours in bright/dark areas for achieving a nicer edge impression"
    IDC_ZONE_BVOPTHRESHOLD  "Change the amount of b-frames in this zone. Recommended values are between -20 (almost no b-vops) and 30 (many b-vops).\nThe hardcoded maximum in profile/level will never be exceeded"
    IDC_LEVEL_PROFILE       "Restrict the usage of MPEG-4 tools and limit the encoded bitrate such that bitstreams are compatible with hardware decoders."
    IDC_LEVEL_WIDTH         "Suggested VOP width (pixels)"
    IDC_LEVEL_HEIGHT        "Suggested VOP height (pixels)"
END

STRINGTABLE DISCARDABLE 
BEGIN
    IDC_CARTOON             "Enables special motion estimation features for cartoons/anime."
    IDC_OVERFLOW_CONTROL_STRENGTH 
                            "0=Default from core (let xvid decide). Else overflow payback percent per frame. Higher value will meet target file size better, but will also spoil quantizer distribution more."
    IDC_ASPECT_RATIO        "Display aspect ratio is used to scale the video on playback/anamorphic encoding).\n\nDefault 1:1 for no scaling necessary."
END

STRINGTABLE DISCARDABLE 
BEGIN
    IDC_PROFILE_PROFILE     "Restrict the usage of MPEG-4 tools and limit the encoded bitrate such that bitstreams are compatible with hardware decoders.\n\nWhen a DXN profile is selected, the following constraints take effect:\n* force 1:1 picture aspect ratio\n* disable bvops if interlaced encoding is enabled\n* force a maximum of 1 consecutive bvop for the portable and HT profiles, and a maximum of 2 consecutive bvops for the HD profile\n* always write divx5 user data string into the bitstream\n* force packed bitstream"
    IDC_BVOP_THRESHOLD      "Change the amount of b-frames in this zone. Recommended values are between -20 (almost no b-vops) and 30 (many b-vops).\nThe hardcoded maximum in profile/level will never be exceeded"
END

STRINGTABLE DISCARDABLE 
BEGIN
    IDC_LEVEL_FPS           "Suggested VOP rate (frames-per-second)"
    IDC_LEVEL_VMV           "Video Memory Verifier (VMV):\n\nThe maximum number of macroblocks permitted per VOP."
    IDC_LEVEL_VCV           "Video Complexity Verifier (VCV):\nThe maximum macroblocks decoded per second."
    IDC_LEVEL_VBV           "Video Buffer Verifier (VBV):\n\nThe maximum size of the video decoder buffer. The encoded bitstream but not overflow or underflow this buffer."
    IDC_LEVEL_BITRATE       "Maximum instantaneous bitrate."
    IDC_BITRATE             "The target AVI bitrate, file size, or quantizer."
    IDC_TRELLISQUANT        "Advanced, high quality quantization mode"
    IDC_BITRATE_S           "Toggle between quantizer, target bitrate, target file size"
END

STRINGTABLE DISCARDABLE 
BEGIN
    IDC_TURBO               "Faster motion estimation for b-frames and quarterpel"
    IDC_BITRATE_TSIZE       "Target file or media size"
END

STRINGTABLE DISCARDABLE 
BEGIN
    IDC_BITRATE_SSIZE       "The file size of subtitles or other data files"
    IDC_BITRATE_SSELECT     "Select file size from existing subtitle file"
    IDC_BITRATE_COVERHEAD   "Calculated container format overhead (kbytes)."
    IDC_BITRATE_VRATE       "Calculated average video bitrate (kilobits-per-second)\nNote that ""desired bitrate"" setting in main window includes AVI overhead, so will be larger than this value."
    IDC_BITRATE_VSIZE       "Calculated video size.\nNote that ""desired filesize"" setting in main window includes AVI overhead, so will be larger than this value."
    IDC_BITRATE_ARATE       "Audio bitrate (kilobits-per-second)"
    IDC_BITRATE_ASELECT     "Select file size from existing audio file"
END

STRINGTABLE DISCARDABLE 
BEGIN
    IDC_CLOSEDGOV           "Closes every group-of-pictures before encoding new keyframe."
    IDC_ZONE_WEIGHT         "Change quality of this zone relative to other zones. It's not recommended to go below 0,2"
    IDC_ZONE_QUANT          "Fix this zone's quality to desired quant"
    IDC_ZONE_GREYSCALE      "Don't code colour information in this zone. You also have to force a keyframe, or old colour information will stay"
END

STRINGTABLE DISCARDABLE 
BEGIN
    IDC_ZONE_FORCEIVOP      "Force a keyframe at the beginnig of the zone"
END

#endif    // Neutral resources
/////////////////////////////////////////////////////////////////////////////



#ifndef APSTUDIO_INVOKED
/////////////////////////////////////////////////////////////////////////////
//
// Generated from the TEXTINCLUDE 3 resource.
//


/////////////////////////////////////////////////////////////////////////////
#endif    // not APSTUDIO_INVOKED

