//{{NO_DEPENDENCIES}}
// Microsoft Developer Studio generated include file.
// Used by resource.rc
//
#define IDC_LICENSE                     3
#define IDD_MAIN                        100
#define IDD_GLOBAL                      101
#define IDD_QUANT                       102
#define IDD_2PASS                       103
#define IDD_RC_2PASS2                   104
#define IDD_CREDITS                     105
#define IDD_CPU                         106
#define IDD_ENC                         107
#define IDD_QUANTMATRIX                 108
#define IDD_ABOUT                       109
#define IDD_2PASSALT                    110
#define IDD_RC_2PASS2_ALT               111
#define IDD_POSTPROC                    112
#define IDD_RC_CBR                      113
#define IDD_PROFILE                     114
#define IDD_RC_2PASS1                   116
#define IDD_MOTION                      117
#define IDD_TOOLS                       120
#define IDD_ZONE                        121
#define IDD_LEVEL                       122
#define IDD_DEC                         124
#define IDD_STATUS                      125
#define IDD_AR                          126
#define IDD_BITRATE                     127
#define IDD_LICENSE                     129
#define IDR_GPL                         131
#define IDD_COMMON                      132
#define WIZ_CYDLG                       140
#define PROP_MIN_CYDLG                  144
#define PROP_SM_CYDLG                   188
#define PROP_MIN_CXDLG                  212
#define PROP_MED_CYDLG                  215
#define PROP_LG_CYDLG                   218
#define PROP_MED_CXDLG                  227
#define PROP_LG_CXDLG                   252
#define WIZ_CXDLG                       276
#define IDC_MODE                        1000
#define IDC_SLIDER                      1001
#define IDC_QUALITY                     1002
#define IDC_QUALITY_ADV                 1005
#define IDC_OTHER                       1006
#define IDC_DEFAULTS                    1007
#define IDC_PROFILE                     1008
#define IDC_MOTION                      1009
#define IDC_QUANTTYPE                   1010
#define IDC_FOURCC                      1011
#define IDC_MAXKEY                      1012
#define IDC_LUMMASK                     1013
#define IDC_MINIQUANT                   1014
#define IDC_MAXIQUANT                   1015
#define IDC_MINPQUANT                   1016
#define IDC_MAXPQUANT                   1017
#define IDC_MINBQUANT                   1018
#define IDC_MAXBQUANT                   1019
#define IDC_QUANTMATRIX                 1020
#define IDC_KFBOOST                     1021
#define IDC_MINKEY                      1022
#define IDC_DISCARD1PASS                1023
#define IDC_DUMMY2PASS                  1024
#define IDC_CURVECOMPH                  1025
#define IDC_CURVECOMPL                  1026
#define IDC_PAYBACK                     1027
#define IDC_PAYBACKBIAS                 1028
#define IDC_PAYBACKPROP                 1029
#define IDC_STATS1                      1030
#define IDC_STATS2                      1031
#define IDC_STATS_BROWSE                1032
#define IDC_STATS2_BROWSE               1033
#define IDC_CREDITS_START               1034
#define IDC_CREDITS_START_BEGIN         1035
#define IDC_CREDITS_START_END           1036
#define IDC_CREDITS_END                 1037
#define IDC_CREDITS_END_START           1038
#define IDC_CREDITS_END_BEGIN           1039
#define IDC_CREDITS_END_END             1040
#define IDC_CREDITS_RATE_RADIO          1041
#define IDC_CREDITS_RATE                1042
#define IDC_CREDITS_QUANT_RADIO         1043
#define IDC_CREDITS_QUANTI              1044
#define IDC_CREDITS_QUANT_STATIC        1045
#define IDC_CREDITS_QUANTP              1046
#define IDC_CREDITS_SIZE_RADIO          1047
#define IDC_CREDITS_START_SIZE          1048
#define IDC_CREDITS_END_STATIC          1049
#define IDC_CREDITS_END_SIZE            1050
#define IDC_CPU_AUTO                    1051
#define IDC_CPU_FORCE                   1052
#define IDC_CPU_MMX                     1053
#define IDC_CPU_MMXEXT                  1054
#define IDC_CPU_SSE                     1055
#define IDC_CPU_SSE2                    1056
#define IDC_CPU_3DNOW                   1057
#define IDC_CPU_3DNOWEXT                1058
#define IDC_LOAD                        1059
#define IDC_SAVE                        1060
#define IDC_WEBSITE                     1061
#define IDC_BUILD                       1062
#define IDC_CORE                        1063
#define IDC_QINTRA00                    1064
#define IDC_QINTRA01                    1065
#define IDC_QINTRA02                    1066
#define IDC_QINTRA03                    1067
#define IDC_QINTRA04                    1068
#define IDC_QINTRA05                    1069
#define IDC_QINTRA06                    1070
#define IDC_QINTRA07                    1071
#define IDC_QINTRA08                    1072
#define IDC_QINTRA09                    1073
#define IDC_QINTRA10                    1074
#define IDC_QINTRA11                    1075
#define IDC_QINTRA12                    1076
#define IDC_QINTRA13                    1077
#define IDC_QINTRA14                    1078
#define IDC_QINTRA15                    1079
#define IDC_QINTRA16                    1080
#define IDC_QINTRA17                    1081
#define IDC_QINTRA18                    1082
#define IDC_QINTRA19                    1083
#define IDC_QINTRA20                    1084
#define IDC_QINTRA21                    1085
#define IDC_QINTRA22                    1086
#define IDC_QINTRA23                    1087
#define IDC_QINTRA24                    1088
#define IDC_QINTRA25                    1089
#define IDC_QINTRA26                    1090
#define IDC_QINTRA27                    1091
#define IDC_QINTRA28                    1092
#define IDC_QINTRA29                    1093
#define IDC_QINTRA30                    1094
#define IDC_QINTRA31                    1095
#define IDC_QINTRA32                    1096
#define IDC_QINTRA33                    1097
#define IDC_QINTRA34                    1098
#define IDC_QINTRA35                    1099
#define IDC_QINTRA36                    1100
#define IDC_QINTRA37                    1101
#define IDC_QINTRA38                    1102
#define IDC_QINTRA39                    1103
#define IDC_QINTRA40                    1104
#define IDC_QINTRA41                    1105
#define IDC_QINTRA42                    1106
#define IDC_QINTRA43                    1107
#define IDC_QINTRA44                    1108
#define IDC_QINTRA45                    1109
#define IDC_QINTRA46                    1110
#define IDC_QINTRA47                    1111
#define IDC_QINTRA48                    1112
#define IDC_QINTRA49                    1113
#define IDC_QINTRA50                    1114
#define IDC_QINTRA51                    1115
#define IDC_QINTRA52                    1116
#define IDC_QINTRA53                    1117
#define IDC_QINTRA54                    1118
#define IDC_QINTRA55                    1119
#define IDC_QINTRA56                    1120
#define IDC_QINTRA57                    1121
#define IDC_QINTRA58                    1122
#define IDC_QINTRA59                    1123
#define IDC_QINTRA60                    1124
#define IDC_QINTRA61                    1125
#define IDC_QINTRA62                    1126
#define IDC_QINTRA63                    1127
#define IDC_QINTER00                    1128
#define IDC_QINTER01                    1129
#define IDC_QINTER02                    1130
#define IDC_QINTER03                    1131
#define IDC_QINTER04                    1132
#define IDC_QINTER05                    1133
#define IDC_QINTER06                    1134
#define IDC_QINTER07                    1135
#define IDC_QINTER08                    1136
#define IDC_QINTER09                    1137
#define IDC_QINTER10                    1138
#define IDC_QINTER11                    1139
#define IDC_QINTER12                    1140
#define IDC_QINTER13                    1141
#define IDC_QINTER14                    1142
#define IDC_QINTER15                    1143
#define IDC_QINTER16                    1144
#define IDC_QINTER17                    1145
#define IDC_QINTER18                    1146
#define IDC_QINTER19                    1147
#define IDC_QINTER20                    1148
#define IDC_QINTER21                    1149
#define IDC_QINTER22                    1150
#define IDC_QINTER23                    1151
#define IDC_QINTER24                    1152
#define IDC_QINTER25                    1153
#define IDC_QINTER26                    1154
#define IDC_QINTER27                    1155
#define IDC_QINTER28                    1156
#define IDC_QINTER29                    1157
#define IDC_QINTER30                    1158
#define IDC_QINTER31                    1159
#define IDC_QINTER32                    1160
#define IDC_QINTER33                    1161
#define IDC_QINTER34                    1162
#define IDC_QINTER35                    1163
#define IDC_QINTER36                    1164
#define IDC_QINTER37                    1165
#define IDC_QINTER38                    1166
#define IDC_QINTER39                    1167
#define IDC_QINTER40                    1168
#define IDC_QINTER41                    1169
#define IDC_QINTER42                    1170
#define IDC_QINTER43                    1171
#define IDC_QINTER44                    1172
#define IDC_QINTER45                    1173
#define IDC_QINTER46                    1174
#define IDC_QINTER47                    1175
#define IDC_QINTER48                    1176
#define IDC_QINTER49                    1177
#define IDC_QINTER50                    1178
#define IDC_QINTER51                    1179
#define IDC_QINTER52                    1180
#define IDC_QINTER53                    1181
#define IDC_QINTER54                    1182
#define IDC_QINTER55                    1183
#define IDC_QINTER56                    1184
#define IDC_QINTER57                    1185
#define IDC_QINTER58                    1186
#define IDC_QINTER59                    1187
#define IDC_QINTER60                    1188
#define IDC_QINTER61                    1189
#define IDC_QINTER62                    1190
#define IDC_QINTER63                    1191
#define IDC_USEALT                      1193
#define IDC_USEAUTO                     1194
#define IDC_AUTOSTR                     1195
#define IDC_USEAUTOBONUS                1196
#define IDC_BONUSBIAS                   1197
#define IDC_CURVETYPE                   1198
#define IDC_ALTCURVEHIGH                1199
#define IDC_ALTCURVELOW                 1201
#define IDC_MINQUAL                     1202
#define IDC_INTERLACING                 1203
#define IDC_OVERDEG                     1204
#define IDC_MAXBFRAMES                  1205
#define IDC_HINTFILE                    1206
#define IDC_BQUANTRATIO                 1207
#define IDC_HINT_BROWSE                 1208
#define IDC_HINTEDME                    1209
#define IDC_OVERIMP                     1210
#define IDC_MAXBITRATE                  1211
#define IDC_CBR_REACTIONDELAY           1212
#define IDC_CBR_AVERAGINGPERIOD         1213
#define IDC_CBR_BUFFER                  1214
#define IDC_PACKED                      1215
#define IDC_BSTATIC1                    1216
#define IDC_MAXBFRAMES_S                1217
#define IDC_BQUANTRATIO_S               1218
#define IDC_DX50BVOP                    1219
#define IDC_DEBUG                       1220
#define IDC_NUMTHREADS                  1221
#define IDC_NUMTHREADS_STATIC           1223
#define IDC_FRAMEDROP                   1224
#define IDC_FRAMEDROP_STATIC            1225
#define IDC_KFTRESHOLD                  1226
#define IDC_KFREDUCTION                 1227
#define IDC_GREYSCALE                   1228
#define IDC_CREDITS_GREYSCALE           1229
#define IDC_SPECIAL_BUILD               1230
#define IDC_GMC                         1231
#define IDC_QPEL                        1232
#define IDC_CHROMAME                    1233
#define IDC_BQUANTOFFSET                1234
#define IDC_DEBLOCK_Y                   1235
#define IDC_DEBLOCK_UV                  1236
#define IDC_REDUCED                     1237
#define IDC_VHQ                         1238
#define IDC_CHROMA_OPT                  1239
#define IDC_STATS                       1240
#define IDC_MODE_ADV                    1241
#define IDC_PROFILE_ADV                 1242
#define IDC_ADD                         1243
#define IDC_PROFILE_BVOP                1244
#define IDC_REMOVE                      1245
#define IDC_EDIT                        1246
#define IDC_PROFILE_MPEGQUANT           1247
#define IDC_BITRATE_CALC                1248
#define IDC_PROFILE_INTERLACE           1249
#define IDC_PROFILE_QPEL                1250
#define IDC_PROFILE_GMC                 1251
#define IDC_PROFILE_REDUCED             1252
#define IDC_PROFILE_WIDTH               1253
#define IDC_PROFILE_HEIGHT              1254
#define IDC_PROFILE_FPS                 1255
#define IDC_PROFILE_BITRATE             1256
#define IDC_PROFILE_VBV                 1257
#define IDC_PROFILE_PROFILE             1258
#define IDC_PROFILE_VCV                 1259
#define IDC_PROFILE_VMV                 1260
#define IDC_BVOP                        1261
#define IDC_BVOP_THRESHOLD              1262
#define IDC_BVOP_THRESHOLD_S            1263
#define IDC_BQUANTOFFSET_S              1265
#define IDC_CLOSEDGOV                   1266
#define IDC_ZONES                       1267
#define IDC_ZONE_FRAME                  1268
#define IDC_ZONE_VALUE                  1269
#define IDC_ZONE_BITRATE                1270
#define IDC_ZONE_WEIGHT                 1274
#define IDC_ZONE_SLIDER                 1275
#define IDC_ZONE_FETCH                  1276
#define IDC_ZONE_MODE                   1277
#define IDC_ZONE_QUANT                  1278
#define IDC_ZONE_GREYSCALE              1279
#define IDC_ZONE_CHROMAOPT              1280
#define IDC_ZONE_BVOPTHRESHOLD_ENABLE   1281
#define IDC_ZONE_BVOPTHRESHOLD          1282
#define IDC_ZONE_MODE_WEIGHT            1285
#define IDC_ZONE_MODE_QUANT             1287
#define IDC_QUANTTYPE_S                 1290
#define IDC_ZONE_MAX                    1291
#define IDC_ZONE_MIN                    1292
#define IDC_LEVEL_PROFILE               1293
#define IDC_LEVEL_WIDTH                 1294
#define IDC_LEVEL_HEIGHT                1295
#define IDC_LEVEL_FPS                   1296
#define IDC_LEVEL_VMV                   1297
#define IDC_LEVEL_VCV                   1298
#define IDC_LEVEL_VBV                   1299
#define IDC_LEVEL_BITRATE               1300
#define IDC_BITRATE                     1302
#define IDC_LEVEL_PEAKRATE              1302
#define IDC_BITRATE_MIN                 1303
#define IDC_BITRATE_MAX                 1304
#define IDC_ZONE_BVOPTHRESHOLD_S        1305
#define IDC_TRELLISQUANT                1306
#define IDC_BITRATE_S                   1307
#define IDC_VOPDEBUG                    1308
#define IDC_STATUS_GRAPH                1311
#define IDC_STATUS_IQ_MIN               1312
#define IDC_STATUS_IQ_MAX               1313
#define IDC_STATUS_PQ_MIN               1314
#define IDC_STATUS_PQ_MAX               1315
#define IDC_STATUS_BQ_MIN               1316
#define IDC_STATUS_BQ_MAX               1317
#define IDC_STATUS_Q_MIN                1318
#define IDC_STATUS_Q_MAX                1319
#define IDC_STATUS_IL_MIN               1320
#define IDC_STATUS_IL_MAX               1321
#define IDC_STATUS_IL_TOT               1322
#define IDC_STATUS_PL_MIN               1323
#define IDC_STATUS_PL_MAX               1324
#define IDC_STATUS_PL_TOT               1325
#define IDC_STATUS_BL_MIN               1326
#define IDC_STATUS_BL_MAX               1327
#define IDC_STATUS_BL_TOT               1328
#define IDC_STATUS_L_MIN                1329
#define IDC_STATUS_L_MAX                1330
#define IDC_STATUS_L_TOT                1331
#define IDC_STATUS_DESTROY              1332
#define IDC_STATUS_KBPS                 1333
#define IDC_STATUS_IL_AVG               1334
#define IDC_DISPLAY_STATUS              1335
#define IDC_STATUS_PL_AVG               1336
#define IDC_STATUS_BL_AVG               1337
#define IDC_FULL1PASS                   1338
#define IDC_DEC_DY                      1339
#define IDC_STATUS_L_AVG                1340
#define IDC_ZONE_FORCEIVOP              1341
#define IDC_STATUS_IQ_AVG               1342
#define IDC_STATUS_I_NUM                1343
#define IDC_STATUS_P_NUM                1344
#define IDC_STATUS_B_NUM                1345
#define IDC_STATUS_NUM                  1346
#define IDC_CARTOON                     1347
#define IDC_STATUS_PQ_AVG               1348
#define IDC_OVERFLOW_CONTROL_STRENGTH   1349
#define IDC_STATUS_BQ_AVG               1350
#define IDC_ASPECT_RATIO                1351
#define IDC_STATUS_Q_AVG                1352
#define IDC_AR                          1353
#define IDC_PAR                         1354
#define IDC_PARX                        1355
#define IDC_PARY                        1356
#define IDC_ARX                         1357
#define IDC_ARY                         1358
#define IDC_AR_DEFAULT                  1364
#define IDC_AR_4_3                      1365
#define IDC_AR_16_9                     1366
#define IDC_AR_235_100                  1368
#define IDC_TURBO                       1369
#define IDC_DEC_DUV                     1370
#define IDC_DEC_DR                      1371
#define IDC_DEC_DRY                     1371
#define IDC_DEBUGOUTPUT                 1372
#define IDC_DEC_DRUV                    1372
#define IDC_SHOWINTERNALS               1373
#define IDC_DEC_FE                      1374
#define IDC_BITRATE_TSIZE               1375
#define IDC_BITRATE_SSIZE               1376
#define IDC_BITRATE_SSELECT             1377
#define IDC_BITRATE_CFORMAT             1378
#define IDC_BITRATE_COVERHEAD           1379
#define IDC_BITRATE_HOURS               1380
#define IDC_BITRATE_MINUTES             1381
#define IDC_BITRATE_SECONDS             1382
#define IDC_BITRATE_FPS                 1383
#define IDC_BITRATE_VRATE               1384
#define IDC_BITRATE_VSIZE               1385
#define IDC_BITRATE_AFORMAT             1386
#define IDC_BITRATE_AMODE_RATE          1387
#define IDC_BITRATE_AMODE_SIZE          1388
#define IDC_BITRATE_ARATE               1389
#define IDC_BITRATE_ASIZE               1390
#define IDC_BITRATE_ASELECT             1391
#define IDC_BITRATE_ADV                 1392
#define IDC_DEC_BRIGHTNESS              1393
#define IDC_TFF                         1394
#define IDC_VHQ_BFRAME                  1395
#define IDC_LICENSE_TEXT                1396
#define IDC_LEVEL_PEAKRATE_S            1397
#define IDC_LEVEL_BITRATE_S             1398
#define IDC_LEVEL_VBV_S                 1399
#define IDC_LEVEL_VBV_G                 1400
#define IDC_LEVEL_DIM_S                 1401
#define IDC_LEVEL_VMV_S                 1402
#define IDC_LEVEL_VCV_S                 1403
#define IDC_LEVEL_LEVEL_G               1404

// Next default values for new objects
// 
#ifdef APSTUDIO_INVOKED
#ifndef APSTUDIO_READONLY_SYMBOLS
#define _APS_NEXT_RESOURCE_VALUE        133
#define _APS_NEXT_COMMAND_VALUE         40001
#define _APS_NEXT_CONTROL_VALUE         1405
#define _APS_NEXT_SYMED_VALUE           101
#endif
#endif
